/*
 * Copyright (c) 2018 Grandstream Networks, Inc.
 * Jianwen Li <jwli@grandstream.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/kernel.h>
#include <linux/switch.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/workqueue.h>

#define POE_LIMIT_DELAYMS 100		    /* 100ms */
#define POE_LIMIT_VBUS_RETRYMS 1000		/* 1s */
#define MAX_DELAY_HANDLE 100		    /* 100 * 100 ms = 10s */

static int usb_vbus_status = 1;

extern int mt_usb20_vbus_switch(int on_off);

static int g_poe = -1;
static int g_poe_802_3at = -1;

int max_delay = MAX_DELAY_HANDLE; /* 100 * 100ms = 10s */

static void poe_limit_delay_handle(struct work_struct *work);

static int poe_limit_inited;

static struct workqueue_struct *poe_limit_delay_wq;
static void poe_limit_usbvbus_retry_handle(struct work_struct *work);
static void poe_limit_delay_handle(struct work_struct *work);
static DECLARE_DELAYED_WORK(poe_limit_delay_work, poe_limit_delay_handle);
static DECLARE_DELAYED_WORK(poe_limit_usbvbus_retry, poe_limit_usbvbus_retry_handle);

static struct switch_dev *g_sdev;
static int g_vbus_switch_to = 1;

static void poe_limit_queue_delayed_work(int ms)
{
	if (!poe_limit_inited) {
		printk("POE: poe_limit_queue_delayed_work Bug: PoE not be initialized\n");
		return;
	}

	queue_delayed_work(poe_limit_delay_wq, &poe_limit_delay_work,
			msecs_to_jiffies(ms));
}

static void poe_limit_usbvbus_queue_delayed_work(int ms)
{
	if (!poe_limit_inited) {
		printk("POE: poe_limit_usbvbus_retry_work Bug: PoE not be initialized\n");
		return;
	}

	queue_delayed_work(poe_limit_delay_wq, &poe_limit_usbvbus_retry,
			msecs_to_jiffies(ms));
}

static void poe_limit_usbvbus_retry_handle(struct work_struct *work)
{
	int ret;
	int vbus_on = g_vbus_switch_to;

	ret = mt_usb20_vbus_switch(vbus_on);
	if (ret < 0) {
		printk("POE: vbus switch failed, try again in 1s\n");
		poe_limit_usbvbus_queue_delayed_work(POE_LIMIT_VBUS_RETRYMS);
	}
}

static void poe_limit_delay_handle(struct work_struct *work)
{
	int usb_vbus_switch;
	int poe_status = 0;   /* '0': 12V-DC; '1': PoE; '2': POE+ */
	struct switch_dev *sdev = g_sdev;

	usb_vbus_switch = 1;
	if (-1 == g_poe || -1 == g_poe_802_3at) {
		if (max_delay-- <= 0) {
			printk(KERN_ERR "POE: poe_limit_delay_handle: timeout\n");
			max_delay = MAX_DELAY_HANDLE;
		} else
			poe_limit_queue_delayed_work(POE_LIMIT_DELAYMS);

		return;
	}

	usb_vbus_switch = 1;
	if (g_poe) {
		printk("POE: Power supply status change (%s)\n",
				1 == g_poe_802_3at ? "POE+" : "POE");
		poe_status = 2;
		if (!g_poe_802_3at) {
			printk("Shutdown USB for POE supply\n");
			usb_vbus_switch = 0;
			poe_status = 1;
		}
	} else {
		poe_status = 0;
		printk("POE: Power supply status change (12V-DC)\n");
	}

	if (usb_vbus_switch != usb_vbus_status) {
		printk("POE: USB20 Vbus Switch %s\n",
				usb_vbus_switch == 1 ? "On" : "Off");

		g_vbus_switch_to = usb_vbus_switch;
		poe_limit_usbvbus_queue_delayed_work(POE_LIMIT_VBUS_RETRYMS);
		usb_vbus_status = usb_vbus_switch;
	}

	switch_set_state(sdev, poe_status);
}

int poe_limit_init(struct switch_dev *sdev)
{
	if (!poe_limit_delay_wq)
		poe_limit_delay_wq = create_singlethread_workqueue("poe_limit_delay_wq");

	if (!g_sdev)
		g_sdev = sdev;

	poe_limit_inited = 1;

	printk("POE: status initialized\n");

	return 0;
}

int poe_limit_update(int poe, int poe_802_3at)
{
	if (-1 != poe_802_3at) {
		g_poe_802_3at = poe_802_3at;
	}

	if (-1 != poe) {
		g_poe = poe;
	}

	poe_limit_queue_delayed_work(POE_LIMIT_DELAYMS);

	return 0;
}

int poe_limit_exit(void)
{
	cancel_delayed_work(&poe_limit_delay_work);
	cancel_delayed_work(&poe_limit_usbvbus_retry);
	flush_workqueue(poe_limit_delay_wq);
	destroy_workqueue(poe_limit_delay_wq);

	return 0;
}
