/*
 * Copyright (c) 2017 Grandstream Networks, Inc.
 * By Fanan Chen <fnchen@grandstream.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/time.h>
#include <linux/gpio.h>
#include <linux/input.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>

struct hksw_priv {
	int hook_status;
	int gpio;
	unsigned int irq;
	struct input_dev *input;
	struct delayed_work work;
};

#define	HOOK_DELAY	100
#define	DEBOUNCE_TIME	10

extern int gvc32xx_get_hw_ver_factory(char *hw_ver);

static int hksw_proc_show(struct seq_file *m, void *v)
{
	struct hksw_priv *hook = (struct hksw_priv *)m->private;

	seq_printf(m, "%s", hook->hook_status ? "up" : "down"); 

	return 0;
}

static int hksw_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file,  hksw_proc_show, PDE_DATA(inode));
}

static const struct file_operations hksw_fops = {
	.open = hksw_proc_open,
	.read = seq_read,
	.release = seq_release,
};

static irqreturn_t hook_interrupt(int irq, void *data)
{
	struct hksw_priv *hook = (struct hksw_priv *)data;

	schedule_delayed_work(&hook->work, msecs_to_jiffies(DEBOUNCE_TIME));

	return IRQ_HANDLED;
}

static void hook_work(struct work_struct *work)
{
	struct hksw_priv *hook =
		container_of(work, struct hksw_priv, work.work);
	int status;

	status = gpio_get_value(hook->gpio);
	if (status == hook->hook_status)
		goto exit;
	else
		hook->hook_status = status;

	if (hook->hook_status) {
		input_report_key(hook->input, KEY_KBDILLUMUP, 1);
		udelay(HOOK_DELAY);
		input_report_key(hook->input, KEY_KBDILLUMUP, 0);
	} else {
		input_report_key(hook->input, KEY_KBDILLUMDOWN, 1);
		udelay(HOOK_DELAY);
		input_report_key(hook->input, KEY_KBDILLUMDOWN, 0);
	}

	input_sync(hook->input);
exit:
	return;
}

static int hook_open(struct input_dev *input_dev)
{
	struct hksw_priv *hook = input_get_drvdata(input_dev);

	enable_irq(hook->irq);

	return 0;
}

static void hook_close(struct input_dev *input_dev)
{
	struct hksw_priv *hook = input_get_drvdata(input_dev);

	disable_irq(hook->irq);
}

static int hksw_probe(struct platform_device *pdev)
{
	struct hksw_priv *hook;
	struct proc_dir_entry *ent;
	struct input_dev *input_dev;
	struct device_node *np = pdev->dev.of_node;
	int err = 0;
	char hw_ver[10];

	hook = kzalloc(sizeof(struct hksw_priv), GFP_KERNEL);
	if (!hook)
		return -ENOMEM;

	input_dev = input_allocate_device();
	if (!input_dev)
		goto fail;

	input_dev->name = pdev->name;
	input_dev->phys = "hook-switch/input0";
	input_dev->id.bustype = BUS_HOST;
	input_dev->dev.parent = &pdev->dev;
	input_dev->open = hook_open;
	input_dev->close = hook_close;
	input_set_drvdata(input_dev, hook);

	__set_bit(EV_KEY, input_dev->evbit);
	__set_bit(KEY_KBDILLUMUP, input_dev->keybit);
	__set_bit(KEY_KBDILLUMDOWN, input_dev->keybit);

	INIT_DELAYED_WORK(&hook->work, hook_work);
	hook->input = input_dev;

	gvc32xx_get_hw_ver_factory (hw_ver);
	if (0 == strncmp (hw_ver, "11B", 3))
		hook->gpio = of_get_named_gpio(np, "compatible-gpio", 0);
	else
		hook->gpio = of_get_named_gpio(np, "common-gpio", 0);
	if (!gpio_is_valid(hook->gpio)) {
		err = -EINVAL;
		dev_err(&pdev->dev, "gpio%d is invalid.\n", hook->gpio);
		goto fail_gpio;
	}

	err = devm_gpio_request_one(&pdev->dev, hook->gpio, GPIOF_IN, pdev->name);
	if (err < 0) {
		err = -EFAULT;
		dev_err(&pdev->dev, "request gpio%d failed.\n",  hook->gpio);
		goto fail_gpio;
	}

	hook->irq = gpio_to_irq(hook->gpio);

	err = request_irq(hook->irq, hook_interrupt,
			  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, "hksw", hook);
	if (err) {
		dev_err(&pdev->dev, "request irq failed\n");
		err = -EINVAL;
		goto fail_allocate;
	}

	err = input_register_device(input_dev);
	if (err)
		goto fail_irq;

	hook->hook_status = gpio_get_value(hook->gpio);

	ent = proc_create_data("key_hook", S_IFREG|S_IRUGO, NULL, &hksw_fops, (void *)hook);
	if (!ent) {
		err = -ENODEV;
		goto fail_proc;
	}

	platform_set_drvdata(pdev, hook);
	disable_irq(hook->irq);

	return 0;

fail_proc:
	input_unregister_device(input_dev);
fail_irq:
	free_irq(hook->irq, hook);
fail_allocate:
fail_gpio:
	input_free_device(input_dev);
fail:
	kfree(hook);
	return err;
}

static int hksw_remove(struct platform_device *pdev)
{
	struct hksw_priv *hook = platform_get_drvdata(pdev);

	free_irq(hook->irq, hook);
	input_unregister_device(hook->input);
	kfree(hook);

	return 0;
}

static const struct of_device_id of_hksw_match[] = {
	{.compatible = "grandstream,hookswitch", },
	{},
};

static struct platform_driver hksw_driver = {
	.probe		= hksw_probe,
	.remove		= hksw_remove,
	.driver		= {
		.name	= "hook-switch",
		.owner	= THIS_MODULE,
		.of_match_table = of_hksw_match,
	}
};

module_platform_driver(hksw_driver);

MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.cn>");
MODULE_DESCRIPTION("Grandstream Hook Switch driver");
MODULE_LICENSE("GPL");
