/*
 * vcom_dac5571.c - Support for LCM VCOM control by DAC5571
 *
 * Copyright (C) 2018 Jianwen Li <jwli@grandstream.cn>
 *
 * This file is subject to the terms and conditions of version 2 of
 * the GNU General Public License.  See the file COPYING in the main
 * directory of this archive for more details.
 */

#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/io.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/platform_device.h>

#include <mach/gs_type.h>
#include "../mmc_rw.h"

#define VCOM_DAC_DRV_NAME "vcom_dac5571"
#define VCOM_DAC_VREF_MV 5140
#define VCOM_DEFAULT_VALUE 3700

#define VCOM_DAC_MAX_VALUE 4500
#define VCOM_DAC_MIN_VALUE 3000

#define VCOM_DAC_MAX_VREF 6000
#define VCOM_DAC_MIN_VREF VCOM_DAC_MAX_VALUE

struct vcom_dac5571_data {
	struct i2c_client *client;
	u16 vref_mv;
	u16 dac_mv;
	u16 dac_value;
	u16 saved_to_flash;
	bool powerdown;
	unsigned powerdown_mode;
	struct miscdevice miscdev;
	unsigned long dev_opened;
};

#define VCOM_IOC_MAGIC     'V'
#define VCOM_GET_DAC_DATA   _IOWR(VCOM_IOC_MAGIC, 0, int)
#define VCOM_SET_DAC_DATA   _IOWR(VCOM_IOC_MAGIC, 1, int)
#define VCOM_SAVE_DAC_DATA  _IOWR(VCOM_IOC_MAGIC, 2, int)
struct vcom_user_data {
	u16 dac_mv;
	u16 vref_mv;
	u16 saved_to_flash;
};

#define VCOM_DATA_MAGIC 0xAD566835
struct vcom_flash_data {
	u32 magic;
	u16 dac_mv;
	u16 vref_mv;
	u16 dac_value; /* not used, caculated by dac_mv and vref_mv */
	u16 reserve[64];
};

/* Platform devices */
static struct platform_device *vcom_dac_device;

static int vcom_dac5571_suspend(struct device *dev)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));
	u8 outbuf[2];

	outbuf[0] = (data->powerdown_mode + 1) << 4;
	outbuf[1] = 0;
	data->powerdown = true;

	return i2c_master_send(data->client, outbuf, 2);
}

static int vcom_dac5571_resume(struct device *dev)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));
	u8 outbuf[2];

	/* restore previous DAC value */
	outbuf[0] = (data->dac_value >> 4) & 0xf;
	outbuf[1] = (data->dac_value & 0xf) << 4;
	data->powerdown = false;

	return i2c_master_send(data->client, outbuf, 2);
}

#ifdef CONFIG_PM_SLEEP
static SIMPLE_DEV_PM_OPS(vcom_dac5571_pm_ops, vcom_dac5571_suspend, vcom_dac5571_resume);
#define VCOM_DAC5571_PM_OPS (&vcom_dac5571_pm_ops)
#else
#define VCOM_DAC5571_PM_OPS NULL
#endif

static int vcom_dac5571_set_value(struct i2c_client *client, int val)
{
	u8 outbuf[2];
	int ret;

	if (val >= (1 << 8) || val < 0)
		return -EINVAL;

	outbuf[0] = (val >> 4) & 0xf;
	outbuf[1] = (val & 0xf) << 4;

	ret = i2c_master_send(client, outbuf, 2);
	if (ret < 0)
		return ret;
	else if (ret != 2)
		return -EIO;
	else
		return 0;
}

static ssize_t vcom_dac5571_voltage_store(struct device *dev, struct device_attribute *attr,
	const char *buf, size_t size)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));
	unsigned char val;
	int ret;

	ret = kstrtou8(buf, 10, &val);
	if (ret)
		return -EINVAL;

	ret = vcom_dac5571_set_value(data->client, val);
	if (ret < 0) {
		printk("vcom dac5571 cat not set value %#x\n", val);
		return size;
	}

	if (data->dac_value != val)
		data->saved_to_flash = 0;

	data->dac_value = val;

	return size;
}

static ssize_t
vcom_dac5571_voltage_show(struct device *dev, struct device_attribute *attr,
	char *buf)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));

	sprintf(buf, "%#x\n", data->dac_value);

	return strlen(buf) + 1;
}

static DEVICE_ATTR(vcom_dac5571_voltage, 0644, vcom_dac5571_voltage_show,
	vcom_dac5571_voltage_store);

static const char *const vcom_dac5571_powerdown_modes[] = {
	"1kohm_to_gnd",
	"100kohm_to_gnd",
	"highimpedance_to_gnd"
};

static ssize_t vcom_dac5571_powerdown_mode_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t size)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));
	unsigned char mode;
	int ret;

	ret = kstrtou8(buf, 10, &mode);
	if (ret)
		return -EINVAL;

	data->powerdown_mode = mode;

	return size;
}

static ssize_t vcom_dac5571_powerdown_mode_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(to_i2c_client(dev));

	if (data->powerdown_mode >= 0 &&
			data->powerdown_mode < ARRAY_SIZE(vcom_dac5571_powerdown_modes)) {
		sprintf(buf, "%s\n", vcom_dac5571_powerdown_modes[data->powerdown_mode]);
	} else {
		sprintf(buf, "unknown powerdown mode\n");
	}

	return strlen(buf) + 1;
}

static DEVICE_ATTR(vcom_dac5571_powerdown_mode, 0644, vcom_dac5571_powerdown_mode_show,
	vcom_dac5571_powerdown_mode_store);

static int vcom_dac5571_load_vcomdata(struct vcom_dac5571_data *vdata);
static int vcom_dac5571_store_vcomdata(struct vcom_dac5571_data *vdata);

static int vcom_dac5571_open(struct inode *inode, struct file *file)
{
	struct vcom_dac5571_data *vdata = container_of(file->private_data,
		struct vcom_dac5571_data, miscdev);

	if (test_and_set_bit(0, &vdata->dev_opened))
		return -EBUSY;

	return 0;
}

static int vcom_dac5571_release(struct inode *inode, struct file *file)
{
	struct vcom_dac5571_data *vdata = container_of(file->private_data,
		struct vcom_dac5571_data, miscdev);

	clear_bit(0, &vdata->dev_opened);

	return 0;
}

static unsigned char vcom_dac5571_dac_value_round(int dac_mv, int dac_refmv)
{
	int dac_round;

	dac_round = dac_mv * 2560 / dac_refmv;
	if ((dac_round % 10) >= 5)
		dac_round = dac_round / 10;
	else
		dac_round = dac_round / 10 - 1;

	return dac_round & 0xff;
}

static long vcom_dac5571_ioctl(struct file *file, unsigned int cmd,
	unsigned long arg)
{
	struct vcom_dac5571_data *vdata = container_of(file->private_data,
		struct vcom_dac5571_data, miscdev);
	struct vcom_user_data v_user_data;
	int ret;
	unsigned char dac_value;

	switch (cmd) {
	case VCOM_GET_DAC_DATA:
		v_user_data.dac_mv = vdata->dac_mv;
		v_user_data.vref_mv = vdata->vref_mv;
		v_user_data.saved_to_flash = vdata->saved_to_flash;
		if (copy_to_user((void __user *)arg, &v_user_data, sizeof(struct vcom_user_data)))
			return -EFAULT;

		break;
	case VCOM_SET_DAC_DATA:
		if (copy_from_user(&v_user_data, (void __user *)arg, sizeof(struct vcom_user_data)))
			return -EFAULT;

		if (v_user_data.dac_mv > VCOM_DAC_MAX_VALUE ||
			v_user_data.dac_mv < VCOM_DAC_MIN_VALUE)
			return -ERANGE;

		if (v_user_data.vref_mv > VCOM_DAC_MAX_VREF ||
			v_user_data.vref_mv < VCOM_DAC_MIN_VREF)
			return -ERANGE;

		dac_value = vcom_dac5571_dac_value_round(v_user_data.dac_mv,
				v_user_data.vref_mv);

		ret = vcom_dac5571_set_value(vdata->client, dac_value);
		if (ret < 0) {
			printk("ioctl: vcom dac5571 cat not set value %#x\n", dac_value);
			return -EINVAL;
		}
		if ((vdata->dac_value != dac_value) ||
			vdata->dac_mv != v_user_data.dac_mv ||
			vdata->vref_mv != v_user_data.vref_mv)
			vdata->saved_to_flash = 0;

		vdata->dac_value = dac_value;
		vdata->dac_mv = v_user_data.dac_mv;
		vdata->vref_mv = v_user_data.vref_mv;

		break;
	case VCOM_SAVE_DAC_DATA:
		ret = vcom_dac5571_store_vcomdata(vdata);
		if (ret < 0) {
			printk("ioctl: vcom dac5571 can not save vcom data\n");
			return -EINVAL;
		}

		break;
	default:
		return -ENOTTY;
	}

	return 0;
}

static int vcom_dac5571_load_vcomdata(struct vcom_dac5571_data *vdata)
{
	int rval = 0;
	struct vcom_flash_data *flash_data;
	sector_t off;

	flash_data = (struct vcom_flash_data *)vmalloc(PAGE_SIZE);
	if (!flash_data)
		return (-ENOMEM);

	memset(flash_data, 0, sizeof(struct vcom_flash_data));
	for (off = 0; off < VCOM_SECTOR_SIZE; off++) {
	rval = emmc_read(VCOM_SECTOR_OFFSET + off, 1, (char *)flash_data, 512);
	if (0 == rval)
		break;
		printk("emmc_read updflag sector(%lu) failed, try next\n", off);
	}

	if (VCOM_SECTOR_SIZE == off) {
		vfree(flash_data);
		return -1;
	}

	if (VCOM_DATA_MAGIC != flash_data->magic) {
		vfree(flash_data);
		return -1;
	}

	vdata->dac_mv = flash_data->dac_mv;
	vdata->vref_mv = flash_data->vref_mv;
	vdata->dac_value = flash_data->dac_value;
	vdata->saved_to_flash = 1;

	vfree(flash_data);

	return 0;
}

static int vcom_dac5571_store_vcomdata(struct vcom_dac5571_data *vdata)
{
	int rval = 0;
	struct vcom_flash_data *flash_data;
	sector_t off;

	flash_data = (struct vcom_flash_data *)vmalloc(PAGE_SIZE);
	if (!flash_data)
		return (-ENOMEM);

	memset(flash_data, 0, sizeof(struct vcom_flash_data));
	flash_data->magic = VCOM_DATA_MAGIC;
	flash_data->dac_mv = vdata->dac_mv;
	flash_data->vref_mv = vdata->vref_mv;
	flash_data->dac_value = vdata->dac_value;

	for (off = 0; off < VCOM_SECTOR_SIZE; off++) {
		rval = emmc_write(VCOM_SECTOR_OFFSET + off, 1, (char *)flash_data, 512);
		if (0 == rval)
			break;
		printk("emmc_write vcom sector(%lu) failed, try next\n", off);
	}

	if (VCOM_SECTOR_SIZE == off) {
		vfree(flash_data);
		return -1;
	}

	printk("write vcomdata: magic(%x), dac_mv(%d), vref_mv(%d), dac_value(0x%x)\n",
		flash_data->magic, flash_data->dac_mv,
		flash_data->vref_mv, flash_data->dac_value);

	vdata->saved_to_flash = 1;

	vfree(flash_data);

	return 0;
}

static const struct file_operations vcom_dac_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.open = vcom_dac5571_open,
	.release = vcom_dac5571_release,
	.unlocked_ioctl = vcom_dac5571_ioctl,
};

static int vcom_dac5571_add_miscdev(struct vcom_dac5571_data *vdata)
{
	int err;
	struct miscdevice *miscdev = &vdata->miscdev;

	miscdev->minor = MISC_DYNAMIC_MINOR;
	miscdev->name = "vcom_dac";
	miscdev->fops = &vcom_dac_fops;
	err = misc_register(miscdev);
	if (err) {
		printk(KERN_ERR "Register %s failed\n", miscdev->name);
	}

	return err;
}

static void vcom_dac5571_remove_miscdev(struct vcom_dac5571_data *vdata)
{
	struct miscdevice *miscdev = &vdata->miscdev;

	misc_deregister(miscdev);
}

static int vcom_dac5571_probe(struct i2c_client *client,
	const struct i2c_device_id *id)
{
	struct vcom_dac5571_data *data;
	int err;

	data = kzalloc(sizeof(struct vcom_dac5571_data), GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	i2c_set_clientdata(client, data);
	data->client = client;
	data->vref_mv = VCOM_DAC_VREF_MV;

	data->powerdown = false;
	data->powerdown_mode = 0;

	err = vcom_dac5571_load_vcomdata(data);
	if (err < 0) {
		/* set default vcom voltage  */
		printk("vcom dac5571 can not load vcom data. restore to default value.\n");
		data->dac_mv = VCOM_DEFAULT_VALUE;
		data->vref_mv = VCOM_DAC_VREF_MV;
	}

	data->dac_value = vcom_dac5571_dac_value_round(data->dac_mv, data->vref_mv);

	printk("vcom dac5571 vcom voltage(%dmv), dac ref(%dmv), round dac(0x%x)\n",
		data->dac_mv, data->vref_mv, data->dac_value);

	err = vcom_dac5571_set_value(data->client, data->dac_value);
	if (err < 0) {
		printk("vcom dac5571 cat not set value %#x\n", data->dac_value);
		goto errset;
	}

	/* Add the vcom_dac device */
	vcom_dac_device = platform_device_alloc("vcom_dac", -1);

	if (vcom_dac_device == NULL) {
		pr_err("Unable to create the vcom_dac device\n");
		err = -ENOMEM;
		goto errset;
	}

	err = platform_device_add(vcom_dac_device);
	platform_set_drvdata(vcom_dac_device, data);
	if (err) {
		pr_err("Unable to add the DCON device\n");
		goto errdev;
	}

	err = device_create_file(&vcom_dac_device->dev,
			&dev_attr_vcom_dac5571_powerdown_mode);
	if (err)
		goto errdev;

	err = device_create_file(&vcom_dac_device->dev,
			&dev_attr_vcom_dac5571_voltage);
	if (err)
		goto errfile0;

	err = vcom_dac5571_add_miscdev(data);
	if (err)
		goto errfile1;

	printk("vcom_dac5571_probe done\n");

	return 0;

errfile1:
	device_remove_file(&vcom_dac_device->dev,
			&dev_attr_vcom_dac5571_voltage);

errfile0:
	device_remove_file(&vcom_dac_device->dev,
			&dev_attr_vcom_dac5571_powerdown_mode);

errdev:
	platform_device_unregister(vcom_dac_device);
	vcom_dac_device = NULL;

errset:
	kfree(data);

	return err;
}

static int vcom_dac5571_remove(struct i2c_client *client)
{
	struct vcom_dac5571_data *data = i2c_get_clientdata(client);

	if (vcom_dac_device) {
		vcom_dac5571_remove_miscdev(data);
		device_remove_file(&vcom_dac_device->dev, &dev_attr_vcom_dac5571_voltage);
		device_remove_file(&vcom_dac_device->dev, &dev_attr_vcom_dac5571_powerdown_mode);
		platform_device_unregister(vcom_dac_device);
		vcom_dac_device = NULL;
	}

	kfree(data);

	return 0;
}

static const struct i2c_device_id vcom_dac5571_id[] = {
	{ "vcom_dac5571", 0 },
	{ }
};
/* MODULE_DEVICE_TABLE(i2c, vcom_dac5571_id); */

static const struct of_device_id vcom_dac5571_dt_match[] = {
	{.compatible = "grandstream,vcom_dac5571"},
	{},
};

MODULE_DEVICE_TABLE(of, vcom_dac5571_dt_match);

static struct i2c_driver vcom_dac5571_driver = {
	.driver = {
		.of_match_table = of_match_ptr(vcom_dac5571_dt_match),
		.name	= VCOM_DAC_DRV_NAME,
		.pm	= VCOM_DAC5571_PM_OPS,
	},
	.probe		= vcom_dac5571_probe,
	.remove		= vcom_dac5571_remove,
	.id_table	= vcom_dac5571_id,
};
module_i2c_driver(vcom_dac5571_driver);

MODULE_AUTHOR("Jianwen Li <jwli@grandstream.cn>");
MODULE_DESCRIPTION("VCOM voltage control by 8-bit DAC DAC5571");
MODULE_LICENSE("GPL");
